# coding: utf8
# Copyright (c) Météo France (2017-)
# This software is governed by the CeCILL-C license under French law.
# http://www.cecill.info
""" This module gathers scores computing functions. """

from __future__ import print_function
import numpy as np
import pandas as pd
from scipy.stats import pearsonr, ttest_ind, levene, mannwhitneyu, spearmanr
from datetime import time
import warnings
cimport numpy as np
cimport cython
from libc.math cimport abs, sqrt


cdef double mean1D(double[:] vect):
    """ 1D array mean computing """
    cdef Py_ssize_t vect_len = vect.shape[0]
    cdef double vect_mean = 0
    cdef Py_ssize_t x
    for x in range(vect_len):
        vect_mean += vect[x]
    vect_mean /= vect_len
    return vect_mean


cdef double variance1D(double[:] vect):
    """ 1D array mean computing """
    cdef Py_ssize_t vect_len = vect.shape[0]
    cdef double var_vect = 0
    cdef Py_ssize_t x
    cdef double m_vect = mean1D(vect)

    for x in range(vect_len):
        var_vect += (vect[x] - m_vect)**2
    var_vect /= vect_len

    return var_vect


cpdef double mean_bias(double[:] obs, double[:] sim):
    """ Mean Bias computing between two numpy arrays """
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef double obs_mean = 0
    cdef double sim_mean = 0
    cdef Py_ssize_t x

    for x in range(x_max):
        obs_mean += obs[x]
        sim_mean += sim[x]

    obs_mean /= x_max
    sim_mean /= x_max

    return sim_mean - obs_mean


def mean_bias_Ttest(obs, sim):
    """ Mean Bias computing between two numpy arrays

    This function returns a bias set to 0 when the T-test for the null
    hypothesis that observations and simulations have identical average
    values is not significant (pvalue > 0.05).

    """
    b = mean_bias(obs, sim)
    pvalue = mannwhitneyu(obs, sim)[1]
    if pvalue > 0.05:
        biasT = 0.
    else:
        biasT = b
    return biasT


cpdef double relative_mean_bias(double[:] obs, double[:] sim):
    """ Relative Mean Bias computing between two numpy arrays """
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef double obs_mean = 0
    cdef double sim_mean = 0
    cdef Py_ssize_t x

    for x in range(x_max):
        obs_mean += obs[x]
        sim_mean += sim[x]

    obs_mean /= x_max
    sim_mean /= x_max

    if obs_mean != 0.:
        return (100.*(sim_mean - obs_mean))/obs_mean
    else:
        return 100.


cpdef double fractional_bias(double[:] obs, double[:] sim):
    """ Fractional bias computing """
    cdef double m_obs = mean1D(obs)
    cdef double m_sim = mean1D(sim)
    if (m_sim + m_obs) != 0:
        return 100 * 2 * (m_sim - m_obs) / (m_sim + m_obs)
    else:
        return 0


cpdef double modified_mean_bias(double[:] obs, double[:] sim):
    """ Modified Mean Bias computing """
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef Py_ssize_t x
    cdef double ratio_sum = 0
    cdef double mmb

    for x in range(x_max):
        if (obs[x] + sim[x]) != 0.:
            ratio_sum += (sim[x] - obs[x])/(obs[x] + sim[x])

    mmb = 2*ratio_sum/x_max

    return mmb


cpdef double fge(double[:] obs, double[:] sim):
    """ Fractional Gross Error computing.

    Parameters
    ----------
    obs : numpy 1D array
        Array containing observations values.
    sim : numpy 1D array
        Array containing simulation values.

    Returns
    -------
        Fractional Gross Error.

    """
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef Py_ssize_t x
    cdef double ratio_sum = 0
    cdef double mmb

    for x in range(x_max):
        if (obs[x] + sim[x]) != 0.:
            ratio_sum += abs((sim[x] - obs[x])/(obs[x] + sim[x]))

    mmb = 2*ratio_sum/x_max

    return mmb


cpdef double variances_ratio(double[:] obs, double[:] sim, bint test=False):
    """ Variance ratio and Levene test. """
    cdef double obs_var = variance1D(obs)
    cdef double sim_var
    cdef double ratio
    cdef double ratioT
    if obs_var != 0.:
        sim_var = variance1D(sim)
        ratio = sim_var/obs_var
        if test is True:
            ltest = levene(obs, sim)[1] < 0.05
            ratioT = ratio*ltest + 1*~ltest
            return ratioT
        else:
            return ratio
    else:
        return 0.


cpdef double RMSE(double[:] obs, double[:] sim):
    """ RMSE computing """
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef double mse = 0
    cdef Py_ssize_t x
    for x in range(x_max):
        mse += (sim[x] - obs[x])**2
    mse /= x_max
    return sqrt(mse)


cpdef double CRMSE(double[:] obs, double[:] sim):
    """ CRMSE computing """
    cdef double m_obs = mean1D(obs)
    cdef double m_sim = mean1D(sim)
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef double crmse = 0
    cdef Py_ssize_t x
    for x in range(x_max):
        crmse += ((sim[x]-m_sim)-(obs[x]-m_obs))**2
    crmse /= x_max
    crmse = sqrt(crmse)
    return crmse


cpdef double NMSE(double[:] obs, double[:] sim):
    """ NMSE computing """
    cdef double m_obs = mean1D(obs)
    cdef double m_sim = mean1D(sim)
    cdef double mean_prod = m_sim*m_obs
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef Py_ssize_t x
    cdef double mse = 0
    if mean_prod != 0:
        for x in range(x_max):
            mse += (sim[x] - obs[x])**2
        mse /= x_max
        return mse/mean_prod
    else:
        return 0


def correlation(obs, sim, type='Pearson'):
    """ Correlation computing """
    if np.std(obs) > 0 and np.std(sim) > 0:
        if type is 'Pearson':
            corr, p = pearsonr(obs, sim)
        elif type is 'Spearman':
            corr, p = spearmanr(obs, sim)
    else:
        corr, p = np.nan, np.nan
    if np.isnan(corr) or np.isnan(p):
        corr = np.nan
        p = np.nan
    corrT = corr*(p < 0.05)
    return corr, corrT


cpdef double PearsonR(double[:] obs, double[:] sim):
    """ Pearson Correlation computing """
    cdef double m_obs = mean1D(obs)
    cdef double m_sim = mean1D(sim)
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef Py_ssize_t x
    cdef double std_obs = 0.
    cdef double std_sim = 0.
    cdef double cov = 0.
    for x in range(x_max):
        std_obs += (obs[x]-m_obs)**2
        std_sim += (sim[x]-m_sim)**2
        cov += (sim[x]-m_sim)*(obs[x]-m_obs)
    std_obs = sqrt(std_obs)
    std_sim = sqrt(std_sim)
    if std_obs > 0. and std_sim > 0.:
        return cov/(std_obs*std_sim)
    else:
        return np.nan


cpdef double factor_of_two(double[:] obs, double[:] sim):
    """ Factor of two """
    cdef double cptr = 0.
    cdef double total = 0.
    cdef Py_ssize_t x_max = obs.shape[0]
    cdef Py_ssize_t x
    for x in range(x_max):
        if obs[x] == 0. and sim[x] == 0.:
            cptr += 1
            total += 1
        elif obs[x] != 0:
            total += 1
            if 0.5 <= sim[x] / obs[x] <= 2:
                cptr += 1
    if total > 0:
        return cptr/total
    else:
        return 0.


def percentile(x, perc):
    """ Value corresponding to the selected percentile. """
    # x_sorted = np.sort(x)
    # ind = int(len(x_sorted)*perc)
    # return x_sorted[ind]
    return np.nanpercentile(x, perc*100, axis=0)


def stats(score):
    """ Various stats & tests """
    dic = {'NbObs': 'len(obs)',
           'Bias': 'mean_bias(obs, sim)',
           'MeanBias': 'mean_bias(obs, sim)',
           'Bias+t': 'mean_bias_Ttest(obs, sim)',
           'FracBias': 'fractional_bias(obs, sim)',
           'biaispct': 'relative_mean_bias(obs, sim)',
           'MMB': 'modified_mean_bias(obs, sim)',
           'FGE': 'fge(obs, sim)',
           'Ratio': 'variances_ratio(obs, sim, test=False)',
           'SDratio': 'np.sqrt(variances_ratio(obs, sim, test=False))',
           'Ratio+t': 'variances_ratio(obs, sim, test=True)',
           'RMSE': 'RMSE(obs, sim)',
           'CRMSE': 'CRMSE(obs, sim)',
           'NMSE': 'NMSE(obs, sim)',
           'PearsonR': "PearsonR(obs, sim)",
           'PearsonR+t': "correlation(obs, sim, type='Pearson')[1]",
           'SpearmanR': "correlation(obs, sim, type='Spearman')[0]",
           'SpearmanR+t': "correlation(obs, sim, type='Spearman')[1]",
           'FactOf2': 'factor_of_two(obs, sim)',
           'sim_median': 'np.median(sim)',
           'obs_median': 'np.median(obs)',
           'sim_mean': 'np.mean(sim)',
           'obs_mean': 'np.mean(obs)',
           'sim_std': 'np.std(sim)',
           'obs_std': 'np.std(obs)',
           'obs_percentile': 'percentile(obs, **kwargs)',
           'sim_percentile': 'percentile(sim, **kwargs)'}
    return eval('lambda obs, sim, **kwargs: ' + dic[score])


def stats2D_numpy(obs, sim, score_list, threshold, **kwargs):
    """ Compute scores from 2D numpy arrays along a given axis.

    Parameters
    ----------
    obs : numpy.ndarray
        Array of observations.
    sim : numpy.ndarray
        Array of simulations.
    score_list : list of str
        List of scores to compute.
    threshold : int or float
        Minimal number (if type(threshold) is int) or minimal rate
        (if type(threshold) is float) of data available in both obs
        and sim required per column (if axis == 0) or per
        row (if axis == 1) to compute the scores.

    Returns
    -------
    numpy.ndarray
        Array with one column per score and rows equal to columns
        (if axis == 0) or rows (if axis == 1) of input DataFrames.

    """
    score_funcs = {}
    for sc in score_list:
        score_funcs[sc] = stats(sc)

    res = np.full((obs.shape[1], len(score_list)), np.nan)
    for col in range(obs.shape[1]):
        idx = np.logical_not(np.isnan(obs[:, col]) | np.isnan(sim[:, col]))
        if np.sum(idx) >= threshold:
            obs_without_nan = obs[:, col][idx]
            sim_without_nan = sim[:, col][idx]
            for i, sc in enumerate(score_list):
                res[col, i] = score_funcs[sc](obs_without_nan,
                                              sim_without_nan, **kwargs)
    return res


def stats2D(obs, sim, score_list, axis=0, threshold=0.75, keep_nan=True,
            **kwargs):
    """ Compute scores from 2D DataFrames along a given axis.

    Parameters
    ----------
    obs : pandas.DataFrame
        DataFrame of observations.
    sim : pandas.DataFrame
        DataFrame of simulations.
    score_list : list of str
        List of scores to compute.
    axis : 0 or 1
        Axis along which to compute the scores.
    threshold : int or float
        Minimal number (if type(threshold) is int) or minimal rate
        (if type(threshold) is float) of data available in both obs and
        sim required per column (if axis == 0) or per row (if axis == 1) to
        compute the scores.
    keep_nan : bool
        If True, np.nan are writen when threshold condition is not met.

    Returns
    -------
    pandas.DataFrame
        DataFrame with one column per score and rows equal to columns
        (if axis == 0) or rows (if axis == 1) of input DataFrames.

    """
    if axis == 1:
        obs = obs.T
        sim = sim.T
    res = pd.DataFrame()

    # define threshold value if it is given as a ratio
    if type(threshold) is float:
        if threshold >= 1.:
            print("*** Warning: threshold of type float >= 1")
        threshold = threshold*obs.shape[0]

    score_array = stats2D_numpy(obs.values, sim.values, score_list=score_list,
                                threshold=threshold, **kwargs)

    return pd.DataFrame(score_array, index=obs.columns, columns=score_list)


def FTstats(obs, sim, score_list, availability_ratio=0.75, day=0, **kwargs):
    """ Compute forecast time scores for each station.

    Parameters
    ----------
    obs : pandas.DataFrame
        DataFrame with datetime index and one column per station.
    sim : pandas.DataFrame
        DataFrame with datetime index and one column per station.
    score_list : list of str
        List of scores to compute.
    availability_ratio : float
        Minimal rate of data (computed scores per time) available on the
        period required per forecast time to compute the mean scores.
    day : int
        The forecast day corresponding to the simulated data.

    Returns
    -------
    dictionary
        Dictionary with keys corresponding to computed scores, that contains
        pandas.DataFrames with one row per forecast time and one column per
        station. For example, rows will be 24, 25, ..., 47 if day = 1.

    """
    # copy obs and sim variables to avoid modifying the original objects
    obs = obs.copy(deep=True)
    sim = sim.copy(deep=True)

    obs.index = pd.MultiIndex.from_arrays([obs.index.date, obs.index.time],
                                          names=['Date', 'Time'])
    sim.index = pd.MultiIndex.from_arrays([sim.index.date, sim.index.time],
                                          names=['Date', 'Time'])
    res = {}
    for s in score_list:
        res[s] = pd.DataFrame()
    for t in range(24):
        o = obs.xs(time(t), level='Time', axis=0)
        s = sim.xs(time(t), level='Time', axis=0)
        st = stats2D(o, s, score_list, axis=0,
                     threshold=float(availability_ratio),
                     keep_nan=True, **kwargs)
        for s in score_list:
            res[s][t+24*day] = st[s]
    return res


def contingency_table(obs, sim, thr):
    """ Contigency table

    Built from a vector of observations and a vector simulations
    compared to a threshold. The contingency table counts the
    number of values greater than the threshold. For example,
    tab[1, 0] is equal to the number of stations where observations
    are lower than the threshold but simulated values are greater
    than the threshold.

    Parameters
    ----------
    obs : numpy array
        Array containing observations values.
    sim : numpy array
        Array containing simulation values.
    thr : float
        Threshold.

    Returns
    -------
    tab : 3x3 numpy array
        tab[2, 0:1] and tab[0:1, 2] contain
        respectively the sum along the lines and the sum along
        the columns of the contingency table.

    """
    # drop nan
    idx = np.logical_not(np.logical_or(np.isnan(obs), np.isnan(sim)))
    obs_without_nan = obs[idx]
    sim_without_nan = sim[idx]

    # contingency table
    n = len(obs_without_nan)
    tab = np.zeros((3, 3))
    idx_obs = obs_without_nan > thr
    idx_sim = sim_without_nan > thr
    tab[0, 0] = np.sum(np.logical_and(idx_obs, idx_sim))
    tab[1, 0] = np.sum(np.logical_and(np.logical_not(idx_obs), idx_sim))
    tab[0, 1] = np.sum(np.logical_and(idx_obs, np.logical_not(idx_sim)))
    tab[1, 1] = n - np.sum(tab)
    tab[2, 0] = np.sum(tab[:, 0])
    tab[2, 1] = np.sum(tab[:, 1])
    tab[0, 2] = np.sum(tab[0, :])
    tab[1, 2] = np.sum(tab[1, :])
    tab[2, 2] = np.sum(tab[2, :])
    return tab


def exceedancesScores(obs, sim, thr):
    """ Different scores """
    idx_nan = np.logical_or(np.isnan(obs), np.isnan(sim))
    o = obs[~idx_nan]
    s = sim[~idx_nan]
    tab = contingency_table(o, s, thr)
    n = tab[2, 2]
    # Calcul des scores
    resul = {}
    if n != 0:
        # Taux de bonnes previsions
        resul['accuracy'] = (tab[0, 0] + tab[1, 1]) / float(n)
        # Bias score
        if tab[2, 0] * tab[0, 2] != 0:
            resul['bias_score'] = tab[2, 0] / float(tab[0, 2])
        # Success ratio
        if tab[0, 0] * tab[2, 0] != 0:
            resul['success_ratio'] = tab[0, 0] / float(tab[2, 0])
        # probability of detection (Hit rate)
        if tab[0, 0] * tab[0, 2] != 0:
            resul['hit_rate'] = tab[0, 0] / float(tab[0, 2])
        # false alarm ratio
        if tab[1, 0] * tab[2, 0] != 0:
            resul['false_alarm_ratio'] = tab[1, 0] / float(tab[2, 0])
        # probability of false detection = false alarm rate
        if tab[1, 0] * tab[1, 2] != 0:
            resul['prob_false_detect'] = tab[1, 0] / float(tab[1, 2])
        # Threat Score
        if tab[0, 2] + tab[1, 0] != 0 and tab[0, 0] != 0:
            resul['threat_score'] = tab[0, 0] / float(tab[0, 2] + tab[1, 0])
        # Equitable Threat Score
        alea = tab[0, 2] * tab[2, 0] / float(n)
        if tab[0, 2] + tab[1, 0] - alea != 0:
            resul['equitable_ts'] = (tab[0, 0] - alea)/float(tab[0, 2] +
                                                             tab[1, 0] - alea)
        # Peirce Skill Score (Hanssen and Kuipers discriminant)
        if tab[0, 2] * tab[1, 2] != 0:
            resul['peirce_ss'] = ((tab[0, 0] / float(tab[0, 2])) -
                                  (tab[1, 0] / float(tab[1, 2])))
        # Heidke Skill Score
        alea = (tab[0, 2] * tab[2, 0] + tab[1, 2] * tab[2, 1]) / float(n)
        if n - alea != 0:
            resul['heidke_ss'] = ((tab[0, 0] + tab[1, 1] - alea) /
                                  float(n - alea))
        # Rousseau Skill Score
        alea = ((tab[0, 2] / 2. + tab[2, 0] / 2.) ** 2 +
                (tab[1, 2] / 2. + tab[2, 1] / 2.) ** 2) / float(n)
        if n - alea != 0:
            resul['rousseau_ss'] = ((tab[0, 0] + tab[1, 1] - alea) /
                                    float(n - alea))
        # Odds Ratio Skill Score
        if tab[0, 0] * tab[1, 1] * tab[1, 0] * tab[0, 1] != 0:
            th = tab[0, 0] * tab[1, 1] / float(tab[1, 0] * tab[0, 1])
            resul['odds_ratio'] = th
            if th + 1 != 0:
                resul['odds_ratio_ss'] = (th - 1) / (th + 1)
    return resul, tab
